"""Module for handling analytics models."""
from dataclasses import dataclass
from typing import Any, Callable

import numpy as np
from numpy.typing import NDArray

from feltlabs.core import randomness
from feltlabs.core.models.base_model import BaseModel


@dataclass
class Metric:
    """Class for defining different analytics which can be calculated.

    Attributes:
        scale_rand: set true if random values are scaled by sample size
        fit_fn: function for calculating metric on original data
        agg_fun: aggregation function for combining results from different models
        remove_fn: function used for calculate negative of aggregated noise model
        output_fn: function used to calculate final value
    """

    scale_rand: bool
    fit_fn: Callable
    agg_fn: Callable
    remove_fn: Callable
    output_fn: Callable = lambda x, _: x


SUPPORTED_MODELS = {
    "Sum": Metric(
        scale_rand=True,
        fit_fn=lambda x: np.sum(x, axis=0),
        agg_fn=lambda vals, _: np.sum(vals, axis=0),
        remove_fn=lambda rands, _: -np.sum(rands, axis=0),
    ),
    "Mean": Metric(
        scale_rand=False,
        fit_fn=lambda x: np.mean(x, axis=0),
        agg_fn=lambda vals, weights: (weights.T @ vals) / np.sum(weights, axis=0),
        remove_fn=lambda rands, weights: -(weights.T @ rands) / np.sum(weights, axis=0),
    ),
    "Variance": Metric(
        # Calculating variance using sums should be numerically more stable
        scale_rand=False,
        fit_fn=lambda x: np.array(
            [
                np.sum(np.power(x, 2), axis=0),
                np.sum(x, axis=0),
            ]
        ),
        agg_fn=lambda vals, _: np.sum(vals, axis=0),
        remove_fn=lambda rands, _: -np.sum(rands, axis=0),
        output_fn=lambda vals, size: abs(
            (vals[0] - vals[1] ** 2 / sum(size)) / sum(size)
        ),
    ),
    "Std": Metric(
        # Calculating variance using sums should be numerically more stable
        scale_rand=False,
        fit_fn=lambda x: np.array(
            [
                np.sum(np.power(x, 2), axis=0),
                np.sum(x, axis=0),
            ]
        ),
        agg_fn=lambda vals, _: np.sum(vals, axis=0),
        remove_fn=lambda rands, _: -np.sum(rands, axis=0),
        output_fn=lambda vals, size: np.sqrt(
            np.abs((vals[0] - vals[1] ** 2 / sum(size)) / sum(size))
        ),
    ),
}


class Model(BaseModel):
    """Model class for scikit-learn models implementing BaseModel."""

    model_type: str = "analytics"
    model: dict[str, NDArray] = {
        "value": np.array([0]),
    }

    def __init__(self, data: dict):
        """Initialize model calss from data dictionary.

        Args
            data: model loaded from JSON as dict
        """
        if data["model_name"] not in SUPPORTED_MODELS:
            raise Exception("Unsupported model type")

        self.model_name = data["model_name"]
        self.metric = SUPPORTED_MODELS[data["model_name"]]
        self.is_dirty = data.get("is_dirty", False)

        params = data.get("model_params", {})
        self._set_params(params)

        self.sample_size = data.get("sample_size", self.sample_size)
        if self.is_dirty:
            # Substract random models (generated from seeds) from loaded model
            self.remove_noise_models(data.get("seeds", []))

    def _export_data(self) -> dict:
        """Get model data as dictionary for storing (and loading) model.

        Returns:
            dictionary containing model data which should be stored
        """
        return {
            "model_type": self.model_type,
            "model_name": self.model_name,
            "is_dirty": self.is_dirty,
            "model_params": self._get_params(),
            "sample_size": self.sample_size,
        }

    def get_random_models(
        self, seeds: list[int], _min: int = -100, _max: int = 100
    ) -> list[BaseModel]:
        """Generate models with random parameters.

        Args:
            seeds: list of seeds for randomness generation
            _min: minimum value of random number
            _max: maximum value of random number

        Returns:
            Returns copy of model with random variables.
        """
        assert len(seeds) == len(
            self.sample_size
        ), f"Can't generate random models. Num seeds ({len(seeds)}) and sizes ({len(self.sample_size)}) missmatch."

        models = []
        for seed, size in zip(seeds, self.sample_size):
            params = self._get_params()
            new_params = {}
            for param, array in params.items():
                self._set_seed(seed, param)

                new_params[param] = randomness.random_array_copy(array, _min, _max)
                if self.metric.scale_rand:
                    new_params[param] *= max(1, size)

            models.append(self.new_model(new_params))
        return models

    def remove_noise_models(self, seeds: list[int]) -> None:
        """Remove generate and remove random models from current model based on seeds.

        Args:
            seeds: list of seeds used for generating random models
        """
        if len(seeds) == 0 or not self.is_dirty:
            return

        noise_models = self.get_random_models(seeds)

        n_model, *other_models = noise_models
        n_model._agg_models_op(self.metric.remove_fn, other_models)

        self._agg_models_op(self.ops["sum_op"], [n_model])
        # Update sample size, because now we have clean aggregated model
        self.sample_size = [sum(self.sample_size)]
        self.is_dirty = False

    def _get_params(self) -> dict[str, NDArray]:
        """Get dictionary of model parameters.

        Returns:
            dictionary of parameters as name to numpy array
        """
        return self.model

    def _set_params(self, params: dict[str, NDArray]) -> None:
        """Set values of model parameters.

        Args:
            params: dictionary mapping from name of param to numpy array
        """
        self.model = {**self.model, **params}

    def _aggregate(self, models: list[BaseModel]) -> None:
        """Aggregation function on self + list of models.

        Args:
            models: list of models
        """
        if len(models) == 0:
            return
        self._agg_models_op(self.metric.agg_fn, models)

    def _fit(self, X: Any, y: Any) -> None:
        """Fit model on given data.

        Args:
            X: array like training data of shape (n_samples, n_features)
            y: array like target values of shapre (n_samples,)
        """
        self.model["value"] = self.metric.fit_fn(y)

    def predict(self, X: Any) -> Any:
        """Use mode for prediction on given data.

        Args:
            X: array like data used for prediciton of shape (n_samples, n_features)
        """
        print(
            f"{self.model_name} value is {self.metric.output_fn(self.model['value'], self.sample_size)}"
        )
        return self.metric.output_fn(self.model["value"], self.sample_size)



"""Module for importing/exporting tensorflow models to json."""
import os
from typing import Any, cast

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

import numpy as np
import tensorflow as tf
from numpy.typing import NDArray

from feltlabs.core.json_handler import json_load
from feltlabs.core.models.base_model import BaseModel
from feltlabs.core.models.tensorflow.cnn_network import get_cnn_network

# TODO: Handle different array types

# So far include only models from tf.keras.applications
SUPPORTED_MODELS = {
    # TODO: Seems to be some issue with mobile net maybe batch norm layer?
    "MobileNetV2": tf.keras.applications.mobilenet_v2.MobileNetV2,
    "EfficientNetB0": tf.keras.applications.efficientnet.EfficientNetB0,
    "CustomCNN": get_cnn_network,
}


class Model(BaseModel):
    """Model class for tensorflow models implementing BaseModel."""

    # TODO: Document model parameters

    model_type: str = "tensorflow"

    def __init__(self, data: dict):
        """Initialize model class from data dictionary.

        Args
            data: model loaded from JSON as dict
        """
        if data["model_name"] not in SUPPORTED_MODELS:
            raise Exception("Unsupported model type")

        self.model_name = data["model_name"]
        self.is_dirty = data.get("is_dirty", False)

        model_class = SUPPORTED_MODELS[self.model_name]
        self.init_params = data.get("init_params", {})
        self.model = model_class(**self.init_params)

        params = data.get("model_params", {})
        self._set_params(params)

        self.fit_args = data.get("fit_args", {})
        self.sample_size = data.get("sample_size", self.sample_size)

        if self.is_dirty:
            # Subtract random models (generated from seeds) from loaded model
            self.remove_noise_models(data.get("seeds", []))
        # Compile model
        self.model.compile(
            "adam", "sparse_categorical_crossentropy", metrics=["accuracy"]
        )

    def _export_data(self) -> dict:
        """Get model data as dictionary for storing (and loading) model.

        Returns:
            dictionary containing model data which should be stored
        """
        return {
            "model_type": self.model_type,
            "model_name": self.model_name,
            "is_dirty": self.is_dirty,
            "init_params": self.init_params,
            "fit_args": self.fit_args,
            "model_params": {
                **self._get_params(),
            },
            "sample_size": self.sample_size,
        }

    def new_model(self, params: dict[str, NDArray] = {}) -> "BaseModel":
        """Create copy of model and set new parameters.

        Args:
            params: values to set to the new model
        """
        new_model = Model(json_load(self.export_model()))
        new_model._set_params(params)
        return new_model

    def remove_noise_models(self, seeds: list[int]) -> None:
        """Remove generate and remove random models from current model based on seeds.

        Args:
            seeds: list of seeds used for generating random models
        """
        if len(seeds) == 0 or not self.is_dirty:
            return

        noise_models = self.get_random_models(seeds)

        op = lambda x, _: -1 * np.mean(x, axis=0)

        n_model, *other_models = noise_models
        n_model._agg_models_op(op, other_models)

        self._agg_models_op(self.ops["sum_op"], [n_model])
        # Cast parameters to required types
        self._set_params(self._get_params())
        # Update sample size, because now we have clean aggregated model
        self.sample_size = [sum(self.sample_size)]
        self.is_dirty = False

    def _get_params(self) -> dict[str, NDArray]:
        """Get dictionary of model parameters.

        Returns:
            dictionary of parameters as name to numpy array
        """
        weights = cast(list[NDArray], self.model.get_weights())
        return dict(map(lambda x: (str(x[0]), x[1]), enumerate(weights)))

    def _set_params(self, params: dict[str, NDArray]) -> None:
        """Set values of model parameters.

        Args:
            params: dictionary mapping from name of param to numpy array
            type_cast: set true if types should be cast to expected type
        """
        if not params:
            return
        sorted_keys = map(str, sorted(map(int, params.keys())))
        weights = [params[i] for i in sorted_keys]
        self.model.set_weights(weights)

    def _aggregate(self, models: list[BaseModel]) -> None:
        """Aggregation function on self + list of models.

        Args:
            models: list of models
        """
        if len(models) == 0:
            return
        self._agg_models_op(self.ops["mean_op"], models)

    def _fit(self, X: Any, y: Any) -> None:
        """Fit model on given data.

        Args:
            X: array like training data of shape (n_samples, n_features)
            y: array like target values of shape (n_samples,)
        """
        self.model.fit(X, y, **self.fit_args)

    def predict(self, X: Any) -> Any:
        """Use mode for prediction on given data.

        Args:
            X: array like data used for prediction of shape (n_samples, n_features)
        """
        return self.model.predict(X)



"""Testing import/export to json."""
from pathlib import Path
from typing import cast

import numpy as np

from feltlabs.core.models import analytics_model, sklearn_model
from feltlabs.core.storage import load_model

X, y = [[0, 0], [1, 1], [2, 2]], [0, 1, 2]


def test_sklearn_linreg_import_export(tmp_path: Path):
    model_data = {"model_type": "sklearn", "model_name": "LinearRegression"}

    model = sklearn_model.Model(model_data)
    model.fit(X, y)

    file_path = tmp_path / "model.json"

    model.export_model(file_path)
    model_bytes = model.export_model()

    with open(file_path, "rb") as f:
        assert f.read() == model_bytes

    im_model = cast(sklearn_model.Model, load_model(file_path))

    assert np.array_equal(im_model.model.coef_, model.model.coef_)
    assert np.array_equal(im_model.predict(X), model.predict(X))


def test_sklearn_ridge_import_export(tmp_path: Path):
    model_data = {
        "model_type": "sklearn",
        "model_name": "Ridge",
        "init_params": {"alpha": 0.1},
    }

    model = sklearn_model.Model(model_data)
    model.fit(X, y)

    file_path = tmp_path / "model.json"

    model.export_model(file_path)
    model_bytes = model.export_model()

    with open(file_path, "rb") as f:
        assert f.read() == model_bytes

    im_model = cast(sklearn_model.Model, load_model(file_path))

    assert np.array_equal(im_model.model.coef_, model.model.coef_)
    assert np.array_equal(im_model.predict(X), model.predict(X))


def test_analytics_import_export(tmp_path: Path):
    model_data = {"model_type": "analytics", "model_name": "Sum"}

    model = analytics_model.Model(model_data)
    model.fit(X, y)

    file_path = tmp_path / "model.json"

    model.export_model(file_path)
    model_bytes = model.export_model()

    with open(file_path, "rb") as f:
        assert f.read() == model_bytes

    im_model = load_model(file_path)

    assert im_model.__dict__ == model.__dict__
    assert np.array_equal(im_model.predict(X), model.predict(X))



"""Module for handling JSON files with numpy arrays."""
import base64
import json
from io import BytesIO
from typing import Any, Union

import numpy as np
from numpy.typing import NDArray

from feltlabs.typing import FileType


def _numpy_to_str(array: Union[NDArray, np.integer, np.floating]) -> str:
    """Convert numpy array to compressed str."""
    bytes_io = BytesIO()
    np.savez_compressed(bytes_io, arr=array)
    enc = base64.b85encode(bytes_io.getvalue())
    return enc.decode()


def _str_to_numpy(data_str: str) -> NDArray:
    """Load numpy array from compressed str."""
    data = base64.b85decode(data_str)
    bytes_io = BytesIO(data)
    # allow_pickle must be False to prevent code execution
    array = np.load(bytes_io, allow_pickle=False)
    return array["arr"]


class NumpyEncoder(json.JSONEncoder):
    """Special class for handling numpy arrays in JSON."""

    # TODO: Remove uncompressed version? But compressed might be useful for some visualisations
    compress = True

    def default(self, obj: Any) -> Any:
        """Special method for encoding numpy arrays into JSON."""
        if (
            isinstance(obj, np.integer)
            or isinstance(obj, np.floating)
            or isinstance(obj, np.ndarray)
        ):
            return {
                "__numpy__": True,
                "array": _numpy_to_str(obj) if self.compress else obj.tolist(),
                "dtype": str(obj.dtype),
                "compress": self.compress,
            }

        return json.JSONEncoder.default(self, obj)

    @staticmethod
    def decoder_hook(dct: dict) -> Any:
        """Static method for decoding numpy arrays from JSON.

        Args:
            dct: dictionary or list of loaded json

        Returns:
            new dictionary with decoded numpy arrays
        """
        if "__numpy__" in dct:
            if dct.get("compress", False):
                return _str_to_numpy(dct["array"])
            return np.array(dct["array"], dtype=dct["dtype"])
        return dct


def json_dump(obj: Any) -> bytes:
    """Convert object to JSON file represented by bytes."""
    return bytes(json.dumps(obj, cls=NumpyEncoder), "utf-8")


def json_load(file: Union[FileType, dict]) -> Any:
    """Load json file using custom loaded.

    Args:
        file: path to json file containing model produced by FELT labs.

    Returns:
        loaded json object (dictionary)
    """
    if isinstance(file, dict):
        return file

    if isinstance(file, bytes):
        return json.loads(file, object_hook=NumpyEncoder.decoder_hook)

    with open(file, "r") as f:
        return json.load(f, object_hook=NumpyEncoder.decoder_hook)



import re

import numpy as np
import tensorflow as tf


def _get_cnn(layer, cnn):
    """Construct CNN layers based on architecture definition."""
    while cnn:
        layer_name, params = cnn.pop(0).split("-", 1)
        if layer_name == "C":
            filters, k_size, stride, padding = params.split("-")
            layer = tf.keras.layers.Conv2D(
                int(filters), int(k_size), int(stride), padding, activation=tf.nn.relu
            )(layer)
        elif layer_name == "CB":
            filters, k_size, stride, padding = params.split("-")
            layer = tf.keras.layers.Conv2D(
                int(filters), int(k_size), int(stride), padding, use_bias=False
            )(layer)
            layer = tf.keras.layers.BatchNormalization()(layer)
            layer = tf.keras.layers.ReLU()(layer)
        elif layer_name == "M":
            k_size, stride = params.split("-")
            layer = tf.keras.layers.MaxPool2D(int(k_size), int(stride))(layer)
        elif layer_name == "R":
            residual = _get_cnn(layer, params[1:-1].split(","))
            layer = tf.keras.layers.Add()([residual, layer])
        elif layer_name == "F":
            layer = tf.keras.layers.Flatten()(layer)
        elif layer_name == "H":
            h_size = params
            layer = tf.keras.layers.Dense(int(h_size), activation=tf.nn.relu)(layer)
        elif layer_name == "D":
            dropout_rate = params
            layer = tf.keras.layers.Dropout(float(dropout_rate))(layer)
    return layer


def get_cnn_network(
    architecture: str, input_shape: tuple[int, int, int], classes: int
) -> tf.keras.Model:
    """Create CNN network based on specified architecture,

    Args:
        architecture: comma-separated list of layers, longer description below
        input_shape: tuple containing (height, width, channels)
        classes: number of final outputs/classes

    Returns:
        keras model which can be used for training and prediction

    Layers description:
        Architecture is comma-separated list of the following layers:
        - `C-filters-kernel_size-stride-padding`: Add a convolutional layer with ReLU
          activation and specified number of filters, kernel size, stride and padding.
          Example: `C-10-3-1-same`
        - `CB-filters-kernel_size-stride-padding`: Same as `C`, but use batch normalization.
          In detail, start with a convolutional layer without bias and activation,
          then add batch normalization layer, and finally ReLU activation. Example: `CB-10-3-1-same`
        - `M-kernel_size-stride`: Add max pooling with specified size and stride. Example: `M-3-2`
        - `R-[layers]`: Add a residual connection. The `layers` contain a specification
          of at least one convolutional layer (but not a recursive residual connection `R`).
          The input to the specified layers is then added to their output
          (after the ReLU nonlinearity of the last one). Example: `R-[C-16-3-1-same,C-16-3-1-same]`
        - `F`: Flatten inputs. Must appear exactly once in the architecture.
        - `H-hidden_layer_size`: Add a dense layer with ReLU activation and specified size. Example: `H-100`
        - `D-dropout_rate`: Apply dropout with the given dropout rate. Example: `D-0.5`
        We assume that provided architecture is valid, else it will fail.
    """
    inputs = tf.keras.layers.Input(shape=input_shape)
    # Produce the results in variable `hidden`.
    # Split by commas which aren't in brackets []
    cnn = re.split(r"(?<!\[[^[\]]),(?![^[\]]*\])", architecture)
    hidden = _get_cnn(inputs, cnn)
    # Add the final output layer
    outputs = tf.keras.layers.Dense(classes, activation=tf.nn.softmax)(hidden)
    model = tf.keras.Model(inputs=inputs, outputs=outputs)
    # TODO: Test if necessary - should init weights
    model(np.expand_dims(np.random.random(input_shape), axis=0))
    return model



"""Module for loading data and models."""
import csv
import json
import pickle
from pathlib import Path
from typing import Any

import numpy as np
import requests
from numpy.typing import NDArray

from feltlabs.config import AggregationConfig, TrainingConfig
from feltlabs.core.cryptography import decrypt_nacl
from feltlabs.core.ocean import get_datasets
from feltlabs.core.storage import load_model


def _has_csv_header(file: Path) -> bool:
    """Check given CSV file if it contains header."""
    with file.open() as f:
        lines = "".join(f.readline() for i in range(5))
        return csv.Sniffer().has_header(lines)


# TODO: Add model type
def load_data(config: TrainingConfig) -> tuple[NDArray, NDArray]:
    """Load data and return them in (X, y) format."""
    if config.data_type == "test":
        # Demo data for testing
        X = np.random.rand(100, 10)
        y = np.random.rand(100)
        return X, y

    datasets = get_datasets(config)

    X, y = [], []
    for dataset in datasets.values():
        data_type = (
            config.data_type if dataset.data_type == "default" else dataset.data_type
        )

        for file, index in dataset.files:
            if data_type == "csv":
                # Check for header
                has_header = _has_csv_header(file)
                data = np.genfromtxt(file, delimiter=",", skip_header=has_header)
                # Get target column index (using modulo to get be positive index)
                index = config.target_column % data.shape[1]
                X.append(
                    np.concatenate([data[:, :index], data[:, index + 1 :]], axis=1)
                )
                y.append(data[:, index])

            elif data_type == "pickle":
                with file.open("rb") as f:
                    data = pickle.load(f)

                # TODO: Do something with the index (probably tabular/tuple data)
                X.append(data[0])
                y.append(data[-1])

    assert len(y) > 0, "No data loaded"

    return np.concatenate(X, axis=0), np.concatenate(y, axis=0)


def load_models(config: AggregationConfig) -> list[Any]:
    """Load models for aggregation.
    It either download models from urls or load them from input path.
    """
    data_array = []
    if config.download_models:
        with config.custom_data_path.open("r") as f:
            conf = json.load(f)

        for url in conf["model_urls"]:
            res = requests.get(url)
            data_array.append(res.content)

    else:
        # Models passed as dataset
        for dataset in get_datasets(config).values():
            for file_path, _ in dataset.files:
                with open(file_path, "rb") as f:
                    data_array.append(f.read())

    models = []
    for val in data_array:
        data = decrypt_nacl(config.private_key, val)
        models.append(load_model(data))

    return models



"""Test saving and loading JSON files with numpy arrays."""
from pathlib import Path

import numpy as np
from deepdiff import DeepDiff

from feltlabs.core.json_handler import json_dump, json_load

obj = {
    "params": np.random.random((4, 5)),
    "params2": [
        np.ones((2, 5), dtype=np.uint),
        np.ones((3, 5), dtype=np.float32),
        np.random.random((4, 5)),
    ],
    "model_type": {"p1": np.ones((2,))},
    "model_name": "LinearRegression",
}


def test_json_numpy_encoding(tmp_path: Path):
    file = tmp_path / "out.json"

    with file.open("wb") as f:
        f.write(json_dump(obj))

    obj_loaded = json_load(file)
    ## Test for no difference between loaded and original
    assert {} == DeepDiff(obj_loaded, obj)



"""Test training process."""
from pathlib import Path

from feltlabs.algorithm import train
from feltlabs.config import parse_training_args
from feltlabs.core.data import load_data
from feltlabs.core.json_handler import json_dump
from feltlabs.core.storage import load_model
from feltlabs.model import load_model

# data = np.array([[0, 0], [1, 1], [2, 2]]), np.array([0, 1, 2])
# Right now it uses test dataset

model_def = {
    "model_type": "sklearn",
    "model_name": "LinearRegression",
}


def test_training(tmp_path: Path):
    input_folder = tmp_path / "input"
    output_folder = tmp_path / "output"

    input_folder.mkdir()
    output_folder.mkdir()

    # Create custom data file (containing model definition)
    with open(input_folder / "algoCustomData.json", "wb") as f:
        f.write(json_dump(model_def))

    ### Training section ###
    args_str = f"--output_folder {output_folder}"
    args_str += f" --input_folder {input_folder}"
    args_str += f" --solo_training"

    args = parse_training_args(args_str.split())

    # Define extra args with different output folder
    model_name = "model.json"
    model = train.main(args_str.split(), output_name=model_name)

    model = load_model(output_folder / model_name)

    # Predict
    data = load_data(args)
    model.predict(data[0])



"""FELT python package intended for running federated learning on Ocean protocol.

This code is intended to work closely with Ocean protocol. Algorithms from this code
should run on ocean provider. Training local models and aggregating them into global
model.

Entry commands:

```bash
felt-train
felt-aggregate
felt-export
```

## Common Usage

After installing this library you can load models trained using FELT as:
```python
from feltlabs.model import load_model

# Load scikit-learn model
model = load_model("final-model.json")

# Data shape must be: (number_of_samples, number_of_features)
data = [
  [1980, 2, 2, 2, 0, 0],
  [1700, 3, 2, 3, 1, 1],
  [2100, 3, 2, 3, 1, 0],
]

result = model.predict(data)
print(result)
# Use following line for analytics as mean, std...
# result = model.predict(None)
```

### Command: felt-export
You can use `felt-export` for converting trained models into pickle object:
Resulting file will then contain a pickled object of scikit-learn model.

```bash
felt-export --input "final-model-House Prices.json" --output "model.pkl"
```

Then you can use the created file as follows:

```python
import pickle

with open('model.pkl', 'rb') as f:
    model = pickle.load(object, f)
    
# See the above code example for data definition
model.predict(data)
```
"""
import re
from pathlib import Path

from setuptools import find_packages, setup

PATH = Path(__file__).parent.absolute()
DOCLINES = (__doc__ or "").split("\n")

CLASSIFIERS = """\
Intended Audience :: Science/Research
Intended Audience :: Developers
Programming Language :: Python
Programming Language :: Python :: 3
Programming Language :: Python :: 3.8
Programming Language :: Python :: 3.9
Programming Language :: Python :: 3.10
Programming Language :: Python :: 3 :: Only
Topic :: Scientific/Engineering
"""


def parse_requirements(file_name):
    """
    from:
        http://cburgmer.posterous.com/pip-requirementstxt-and-setuppy
    """
    requirements = []
    with open(file_name, "r") as f:
        for line in f:
            if re.match(r"(\s*#)|(\s*$)", line):
                continue
            if re.match(r"\s*-e\s+", line):
                requirements.append(
                    re.sub(r"\s*-e\s+.*#egg=(.*)$", r"\1", line).strip()
                )
            elif re.match(r"\s*-f\s+", line):
                pass
            else:
                requirements.append(line.strip())
    return requirements


requirements = parse_requirements(PATH / "requirements.txt")


setup(
    name="feltlabs",
    version="0.4.3",
    packages=find_packages(),
    maintainer="FELT Labs",
    maintainer_email="info@bretahajek.com",
    description=DOCLINES[0],
    long_description="\n".join(DOCLINES[2:]),
    long_description_content_type="text/markdown",
    keywords=["Federated Learning", "Web3", "Machine Learning"],
    url="https://feltlabs.ai/",
    author="FELT Labs",
    project_urls={
        "Bug Tracker": "https://github.com/FELT-Labs/feltlabs.py/issues",
        "Documentation": "https://docs.feltlabs.ai/",
        "Source Code": "https://github.com/FELT-Labs/feltlabs.py",
    },
    license="GPL-3.0 License",
    classifiers=[_f for _f in CLASSIFIERS.split("\n") if _f],
    platforms=["Windows", "Linux", "Solaris", "Mac OS-X", "Unix"],
    python_requires=">=3.9",
    install_requires=requirements,
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "felt-train = feltlabs.algorithm.train:main",
            "felt-aggregate = feltlabs.algorithm.aggregate:main",
            "felt-export = feltlabs.model:main",
        ],
    },
)



"""Module for loading data provider (Node) config."""
import argparse
import json
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, cast

OUTPUT_FOLDER = Path("/data/outputs")
INPUT_FOLDER = Path("/data/inputs/")
CUSTOM_DATA = "algoCustomData.json"


@dataclass
class OceanConfig:
    input_folder: Path = INPUT_FOLDER
    output_folder: Path = OUTPUT_FOLDER
    custom_data_path: Path = INPUT_FOLDER / CUSTOM_DATA
    custom_data: str = CUSTOM_DATA


@dataclass
class TrainingConfig(OceanConfig):
    aggregation_key: bytes = bytes(0)
    data_type: str = "test"
    seed: int = 42
    target_column: int = -1
    solo_training: bool = False
    experimental: bool = False


@dataclass
class AggregationConfig(OceanConfig):
    private_key: bytes = bytes(0)
    public_key: Optional[bytes] = None
    download_models: bool = False
    min_models: int = 2


def _help_exit(parser, error_msg=None):
    """Print help of parser and quit the script."""
    parser.print_help()
    if error_msg:
        print(f"\nERROR: {error_msg}")
    sys.exit(2)


def _add_ocean_config(config: OceanConfig) -> dict[str, str]:
    """Load json file containing algorithm's custom data and add them to config.

    Args:
        config: ocean config containing output path

    Returns:
        dict representing loaded JSON file
    """
    config.custom_data_path = config.input_folder / config.custom_data
    if not config.custom_data_path.exists():
        return {}

    with config.custom_data_path.open("r") as f:
        return json.load(f)


def _ocean_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Add arguments for parsing ocean path configs."""
    parser.add_argument(
        "--output_folder",
        type=Path,
        default=OUTPUT_FOLDER,
        help="Folder for storing outputs.",
    )
    parser.add_argument(
        "--input_folder",
        type=Path,
        default=INPUT_FOLDER,
        help="Folder containing input data.",
    )
    parser.add_argument(
        "--custom_data",
        type=str,
        default=CUSTOM_DATA,
        help="Name of custom data file",
    )
    return parser


def parse_training_args(args_str: Optional[list[str]] = None) -> TrainingConfig:
    """Parse and partially validate arguments form command line.
    Arguments are parsed from string args_str or command line if args_str is None

    Args:
        args_str: list with string arguments or None if using command line

    Returns:
        Parsed args object
    """
    parser = argparse.ArgumentParser(
        description="""
            Script for training models, possible to execute from command line.
            At least one of the flags --solo_training or --aggregation_key KEY must be set.
        """,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser = _ocean_parser(parser)
    # Fixed algorithm params (set as part of deployed algorithm)
    parser.add_argument(
        "--aggregation_key",
        type=str,
        help="Public key used for encrypting local model for aggregation algorithm.",
    )
    parser.add_argument(
        "--solo_training",
        action="store_true",
        default=False,
        help="If true (flag included), it will run training on single dataset without encryption.",
    )
    parser.add_argument(
        "--experimental",
        action="store_true",
        default=False,
        help="If true, allow usage of experimental models (which might be less secure).",
    )
    # Changable arguments (can be changed via algorithm custom data)
    parser.add_argument(
        "--data_type",
        type=str,
        default="test",
        choices=["test", "csv", "pickle"],
        help="Select type of data. For csv last column is used as Y.",
    )
    parser.add_argument(
        "--target_column",
        type=int,
        default=-1,
        help="Select index of target column.",
    )
    parser.add_argument(
        "--seed",
        type=int,
        default=42,
        help="Seed for randomness generation. It should be different for every run.",
    )
    args = parser.parse_args(args_str)

    conf = _add_ocean_config(cast(OceanConfig, args))
    args.data_type = conf["data_type"] if "data_type" in conf else args.data_type
    args.seed = conf["seed"] if "seed" in conf else args.seed
    args.target_column = (
        conf["target_column"] if "target_column" in conf else args.target_column
    )

    if not args.solo_training and not args.aggregation_key:
        # At least one of solo_training or aggregation_key must be set (else exit)
        _help_exit(
            parser, "At least one of --solo_training, --aggregation_key must be set"
        )

    if not args.solo_training:
        args.aggregation_key = bytes.fromhex(args.aggregation_key)

    return cast(TrainingConfig, args)


def parse_aggregation_args(args_str: Optional[list[str]] = None) -> AggregationConfig:
    """Parse and partially validate arguments form command line.
    Arguments are parsed from string args_str or command line if args_str is None

    Args:
        args_str: list with string arguments or None if using command line

    Returns:
        Parsed args object
    """
    parser = argparse.ArgumentParser(
        description="Script for training models, possible to execute from command line.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser = _ocean_parser(parser)
    # Fixed algorithm params (set as part of deployed algorithm)
    parser.add_argument(
        "--private_key",
        type=str,
        help="Private key specified for aggregation algorithm used for decrypting models.",
    )
    parser.add_argument(
        "--min_models",
        type=int,
        default=2,
        help="Minimum number of models required for aggregation (agg fails otherwise).",
    )
    parser.add_argument(
        "--download_models",
        help="If true (flag included), models will be donwloaded from provided URLs",
        action="store_true",
        default=False,
    )
    # Changable arguments (can be changed via algorithm custom data)
    parser.add_argument(
        "--public_key",
        type=str,
        help="Public key used for encrypting final model for scientis.",
        default=None,
    )
    args = parser.parse_args(args_str)

    conf = _add_ocean_config(cast(OceanConfig, args))
    args.public_key = conf["public_key"] if "public_key" in conf else args.public_key

    if not args.private_key:
        # Private key is required
        _help_exit(parser, "Private key is not defined.")

    args.private_key = bytes.fromhex(args.private_key)
    if args.public_key:
        args.public_key = bytes.fromhex(args.public_key)

    return cast(AggregationConfig, args)
